/*
    ChibiOS/RT - Copyright (C) 2006-2007 Giovanni Di Sirio.

    This file is part of ChibiOS/RT.

    ChibiOS/RT is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 3 of the License, or
    (at your option) any later version.

    ChibiOS/RT is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

/**
 * @file    xtensa/chcore.h
 * @brief   xtensa architecture port macros and structures.
 * @details This file is a template of the system driver macros provided by
 *          a port.
 *
 * @addtogroup XTENSA_CORE
 * @{
 */

#ifndef _CHCORE_H_
#define _CHCORE_H_

#include "ch.h"

/**
 * @brief   Type of a generic xtensa register.
 */
typedef void *regxtensa_t;

/**
 * @name    Port Capabilities and Constants
 * @{
 */
/**
 * @brief   This port supports a realtime counter.
 */
#define PORT_SUPPORTS_RT                FALSE

/**
 * @brief   Natural alignment constant.
 * @note    It is the minimum alignment for pointer-size variables.
 */
#define PORT_NATURAL_ALIGN              sizeof (void *)

/**
 * @brief   Stack alignment constant.
 * @note    It is the alignement required for the stack pointer.
 */
#define PORT_STACK_ALIGN                sizeof (stkalign_t)

/**
 * @brief   Working Areas alignment constant.
 * @note    It is the alignment to be enforced for thread working areas.
 */
#define PORT_WORKING_AREA_ALIGN         PORT_STACK_ALIGN
/** @} */

/**
 * @brief   If enabled allows the idle thread to enter a low power mode.
 */
#ifndef ENABLE_WFI_IDLE
#define ENABLE_WFI_IDLE 0
#endif

/**
 * @brief   Unique macro for the implemented architecture.
 */
#define CH_ARCHITECTURE_XTENSA

/**
 * @brief   Name of the implemented architecture.
 */
#define CH_ARCHITECTURE_NAME "xtensa lx106"

/**
 * @brief   16-bytes stack alignment.
 */
typedef struct {uint32_t stkalign[4];} __attribute__((aligned(16))) stkalign_t;
//typedef uint64_t stkalign_t;

/*
d0 - return register
d0, d1, a0, a1, a7 - caller-save
others - callee-saved
*/

/**
 * @brief   Interrupt saved context.
 * @details This structure represents the stack frame saved during a
 *          preemption-capable interrupt handler.
 * @note    Checked for O2 and Os optimizations only.
 * @note    These registers are caller-save
 * @note    *** Size must be multiply of 16 ***
 */
struct port_extctx {
  regxtensa_t a2;
  regxtensa_t a3;
  regxtensa_t a4;
  regxtensa_t a5;
  regxtensa_t a6;
  regxtensa_t a7;
  regxtensa_t a8;
  regxtensa_t a9;
  regxtensa_t a10;
  regxtensa_t a11;
  regxtensa_t ps;
  regxtensa_t a0;
};
/*
struct port_extctx {
  uint32_t  d0;
  uint32_t  d1;
  uint32_t  a0;
  uint32_t  a1;
  uint32_t  sr;
  uint32_t  pc;
};
*/

/**
 * @brief   System saved context.
 * @details This structure represents the inner stack frame during a context
 *          switching.
 * @note    These registers are callee-saved
 * @note    *** Size must be multiply of 16 ***
 */
struct port_intctx {
  regxtensa_t pad1;
  regxtensa_t pad2;
  regxtensa_t a12;
  regxtensa_t a13;
  regxtensa_t a14;
  regxtensa_t a15;
  regxtensa_t ps;
  regxtensa_t a0;
};
/*
struct port_intctx {
  uint32_t  d2;
  uint32_t  d3;
  uint32_t  d4;
  uint32_t  d5;
  uint32_t  d6;
  uint32_t  d7;
  uint32_t  a2;
  uint32_t  a3;
  uint32_t  a4;
  uint32_t  a5;
  uint32_t  a6;
  uint32_t  pc;
};
*/

/**
 * @brief   Platform dependent part of the @p Thread structure.
 * @details In the ColdFire port this structure just holds a pointer to the
 *          @p intctx structure representing the stack pointer at the time
 *          of the context switch.
 */
struct port_context {
  struct port_intctx *sp;
};

/**
 * @brief   Platform dependent part of the @p chThdInit() API.
 * @details This code usually setup the context switching frame represented
 *          by an @p intctx structure.
 * @note    Initial values for registers d2-d7 and a3-a4 are for test purposes only.
 */
/*
#define PORT_SETUP_CONTEXT(workspace, wsize, pf, arg) {                     \
  tp->p_ctx.sp = (struct port_intctx *)((uint8_t *)workspace +              \
                                   wsize -  sizeof(struct port_intctx));    \
  tp->p_ctx.sp->d2 = 0xD2D2D2D2;                                            \
  tp->p_ctx.sp->d3 = 0xD3D3D3D3;                                            \
  tp->p_ctx.sp->d4 = 0xD4D4D4D4;                                            \
  tp->p_ctx.sp->d5 = 0xD5D5D5D5;                                            \
  tp->p_ctx.sp->d6 = 0xD6D6D6D6;                                            \
  tp->p_ctx.sp->d7 = 0xD7D7D7D7;                                            \
  tp->p_ctx.sp->a3 = 0xA3A3A3A3;                                            \
  tp->p_ctx.sp->a4 = 0xA4A4A4A4;                                            \
  tp->p_ctx.sp->a5 = (uint32_t)pf;                                          \
  tp->p_ctx.sp->a6 = (uint32_t)arg;                                         \
  tp->p_ctx.sp->pc = (uint32_t)_port_thread_start;                          \
}
*/
/**
 * @brief   Platform dependent part of the @p chThdCreateI() API.
 * @details This code usually setup the context switching frame represented
 *          by an @p port_intctx structure.
 */
#define PORT_SETUP_CONTEXT(tp, wbase, wtop, pf, arg) {                      \
  (tp)->ctx.sp = (struct port_intctx *)((uint8_t *)(wtop) -                 \
                                        sizeof (struct port_intctx));       \
  (tp)->ctx.sp->a12 = (regxtensa_t)(pf);                                    \
  (tp)->ctx.sp->a13 = (regxtensa_t)(arg);                                   \
  (tp)->ctx.sp->a14 = (regxtensa_t)(0x14141414);                            \
  (tp)->ctx.sp->a15 = (regxtensa_t)(0x15151515);                            \
  (tp)->ctx.sp->ps = (regxtensa_t)0x2F;                                     \
  (tp)->ctx.sp->a0 = (regxtensa_t)_port_thread_start;                       \
}

/**
 * @brief   Stack size for the system idle thread.
 * @details This size depends on the idle thread implementation, usually
 *          the idle thread should take no more space than those reserved
 *          by @p INT_REQUIRED_STACK.
 */
#ifndef PORT_IDLE_THREAD_STACK_SIZE
#define PORT_IDLE_THREAD_STACK_SIZE 256
#endif

/**
 * @brief   Per-thread stack overhead for interrupts servicing.
 * @details This constant is used in the calculation of the correct working
 *          area size.
 *          This value can be zero on those architecture where there is a
 *          separate interrupt stack and the stack space between @p intctx and
 *          @p extctx is known to be zero.
 */
#ifndef PORT_INT_REQUIRED_STACK
#define PORT_INT_REQUIRED_STACK 128
#endif

/**
 * @brief   Enforces a correct alignment for a stack area size value.
 */
//#define PORT_STACK_ALIGN(n) ((((n) - 1) | (sizeof(stkalign_t) - 1)) + 1)

/**
 * @brief   Computes the thread working area global size.
 */
#define PORT_WA_SIZE(n)     (sizeof(struct port_intctx) +             \
                             sizeof(struct port_extctx) +             \
                             ((size_t)(n) + (size_t)(PORT_INT_REQUIRED_STACK)))

/**
 * @brief   Static working area allocation.
 * @details This macro is used to allocate a static thread working area
 *          aligned as both position and size.
 */
#define PORT_WORKING_AREA(s, n) stkalign_t s[PORT_WA_SIZE(n) / sizeof(stkalign_t)];

/**
 * @brief   IRQ prologue code.
 * @details This macro must be inserted at the start of all IRQ handlers
 *          enabled to invoke system APIs.
 * @note    Currently the ColdFire port prevents the IRQ handlers preemption
 *          by raising the mask into the SR register to 7.
 */
/*
#define PORT_IRQ_PROLOGUE() {                                               \
  asm volatile ("movew    #0x2700, %sr");                                   \
}
*/

/**
 * @brief   IRQ epilogue code.
 * @details This macro must be inserted at the end of all IRQ handlers
 *          enabled to invoke system APIs.
 */
/*
#define PORT_IRQ_EPILOGUE() {                                               \
  if (chSchIsRescRequiredExI())                                             \
    chSchDoRescheduleI();                                                   \
}
*/

/**
 * @brief   IRQ handler function declaration.
 * @note    @p id can be a function name or a vector number depending on the
 *          port implementation.
 */
//#define PORT_IRQ_HANDLER(id) void __attribute__((interrupt_handler)) id(void)
//#define PORT_IRQ_HANDLER(id)

/**
 * @brief   Port-related initialization code.
 * @note    Empty in this port.
 */
#define port_init()

/**
 * @brief   Returns a word encoding the current interrupts status.
 *
 * @return              The interrupts status.
 */
static inline syssts_t port_get_irq_status(void) {

//  return (syssts_t)__get_PRIMASK();
  return (syssts_t)0;
}

/**
 * @brief   Checks the interrupt status.
 *
 * @param[in] sts       the interrupt status word
 *
 * @return              The interrupt status.
 * @retval false        the word specified a disabled interrupts status.
 * @retval true         the word specified an enabled interrupts status.
 */
static inline bool port_irq_enabled(syssts_t sts) {

  return (sts & (syssts_t)1) == (syssts_t)0;
}

/**
 * @brief   Determines the current execution context.
 *
 * @return              The execution context.
 * @retval false        not running in ISR mode.
 * @retval true         running in ISR mode.
 */
static inline bool port_is_isr_context(void) {

//  return (bool)((__get_IPSR() & 0x1FFU) != 0U);
  return (bool)false;
}

#ifndef XTSTR
#define _XTSTR(x)	# x
#define XTSTR(x)	_XTSTR(x)
#endif

#define XTOS_SET_INTLEVEL(intlevel)		({ unsigned __tmp; \
			__asm__ __volatile__(	"rsil	%0, " XTSTR(intlevel) "\n" \
						: "=a" (__tmp) : : "memory" ); \
			__tmp;})

/**
 * @brief   Kernel-lock action.
 * @details Usually this function just disables interrupts but may perform more
 *          actions.
 * @note    In the ColdFire port this is performed by raising the IRQ mask
 *          in the SR register to 7.
 */
//#define port_lock() asm volatile ("move.w    #0x2700, %sr")
#define port_lock()   XTOS_SET_INTLEVEL(15)

/**
 * @brief   Kernel-unlock action.
 * @details Usually this function just disables interrupts but may perform more
 *          actions.
 * @note    In the ColdFire port this is performed by lowering the IRQ mask
 *          in the SR register to 0.
 */
//#define port_unlock() asm volatile ("move.w    #0x2000, %sr")
#define port_unlock() XTOS_SET_INTLEVEL(0)

/**
 * @brief   Kernel-lock action from an interrupt handler.
 * @details This function is invoked before invoking I-class APIs from
 *          interrupt handlers. The implementation is architecture dependent,
 *          in its simplest form it is void.
 * @note    Empty in this port.
 */
#define port_lock_from_isr()

/**
 * @brief   Kernel-unlock action from an interrupt handler.
 * @details This function is invoked after invoking I-class APIs from interrupt
 *          handlers. The implementation is architecture dependent, in its
 *          simplest form it is void.
 * @note    Empty in this port.
 */
#define port_unlock_from_isr()

/**
 * @brief   Disables all the interrupt sources.
 * @note    Of course non maskable interrupt sources are not included.
 */
//#define port_disable() asm volatile ("move.w    #0x2700, %sr")
#define port_disable()    port_lock()

/**
 * @brief   Disables the interrupt sources that are not supposed to preempt
 *          the kernel.
 */
//#define port_suspend() asm volatile ("move.w    #0x2700, %sr")
#define port_suspend()    port_lock()

/**
 * @brief   Enables all the interrupt sources.
 */
//#define port_enable() asm volatile ("move.w    #0x2000, %sr")
#define port_enable()     port_unlock()

/**
 * This port function is implemented as inlined code for performance reasons.
 * @note The port code does not define a low power mode, this macro has to be
 *       defined externally. The default implementation is a "nop", not a
 *       real low power mode.
 */
#if ENABLE_WFI_IDLE != 0
#ifndef port_wait_for_interrupt
#define port_wait_for_interrupt() {                                     \
  asm volatile ("waiti  0");                                            \
}
#endif
#else
#define port_wait_for_interrupt()
#endif

#ifdef __cplusplus
extern "C" {
#endif
  void port_halt(void);
  void port_switch(thread_t *ntp, thread_t *otp);
  void _port_thread_start(void);
#ifdef __cplusplus
}
#endif

#endif /* _CHCORE_H_ */

/** @} */
